.486p
model flat
ideal

; finger size
xMaxFinger=4095
yMaxFinger=4095

; screen size
xMaxScreen=1920
yMaxScreen=1080

extrn DosClose:near
extrn DosCreateThread:near
extrn DosExit:near
extrn DosExitList:near
extrn DosOpen:near
extrn DosSetPriority:near
extrn DosSleep:near
extrn DosSuspendThread:near
extrn DosWrite:near

stack 8192

dataseg
szFinger db '\DEV\$CP3924$',0
szScreen db '\DEV\XSMOUSE$',0

dataseg
sGood0 db 'started.',13,10
sGood1 db 'stopped.',13,10
sGood2 db 'waiting.',13,10
sGood3 db 'polling.',13,10
sGood4 db 'quitting.',13,10
label sGood5 byte

dataseg
sInfo0 db ' opening finger',13,10
sInfo1 db ' opening screen',13,10
sInfo2 db ' closing screen',13,10
sInfo3 db ' closing finger',13,10
label sInfo4 byte

udataseg
ActionTaken dd ?
BytesDone dd ?
fhFinger dd ?
fhScreen dd ?

dataseg
PutMouseEvent dw 1,yMaxScreen/2,xMaxScreen/2,yMaxScreen,xMaxScreen

udataseg
tidFinger dd ?

codeseg
proc MainRoutine c near
arg @@Mod,@@Nul,@@Env,@@Arg
; determine begin of arguments
  cld ; operate foreward scan
  mov ecx,512 ; max scan length
  mov edi,[@@Arg] ; start address
  repne scasb ; find terminator
; process passed arguments
  call ProcessArguments
; show application started message
  call ShowInfo c,offset(sGood0),sGood1-sGood0
; open finger device driver
  call ShowInfo c,offset(sInfo0),sInfo1-sInfo0
  call DosOpen c,offset(szFinger),offset(fhFinger),offset(ActionTaken),0,0,1,0192h,0
  call ShowReturnCode
  jnz NotOpenFingerDriver
; open screen device driver
  call ShowInfo c,offset(sInfo1),sInfo2-sInfo1
  call DosOpen c,offset(szScreen),offset(fhScreen),offset(ActionTaken),0,0,1,0192h,0
  call ShowReturnCode
  jnz NotOpenScreenDriver
; register termination processing
  call DosExitList c,1,offset(ProcessComplete)
; move cursor to center of screen
  call DosWrite c,[fhScreen],offset(PutMouseEvent),10,offset(BytesDone)
; start finger input thread
  call DosCreateThread c,offset(tidFinger),offset(FingerThread),0,2,8192
; show application waiting message
  call ShowInfo c,offset(sGood2),sGood3-sGood2
; hang in here forever
  call DosSleep c,-1
label EndProcess near
; force process complete
  sub eax,eax ; success
  ret ; uses exit list
label NotOpenScreenDriver near
; close finger device driver
  call ShowInfo c,offset(sInfo3),sInfo4-sInfo3
  call DosClose c,[fhFinger]
  call ShowReturnCode
label NotOpenFingerDriver near
; show application stopped message
  call ShowInfo c,offset(sGood1),sGood2-sGood1
; exit the process
  call DosExit c,1,0
endp MainRoutine

codeseg
proc ProcessComplete c near
arg @@ReasonCode
; show application quitting message
  call ShowInfo c,offset(sGood4),sGood5-sGood4
; suspend finger input thread
  call DosSuspendThread c,[tidFinger]
; close screen device driver
  call ShowInfo c,offset(sInfo2),sInfo3-sInfo2
  call DosClose c,[fhScreen]
  call ShowReturnCode
; close finger device driver
  call ShowInfo c,offset(sInfo3),sInfo4-sInfo3
  call DosClose c,[fhFinger]
  call ShowReturnCode
; show application stopped message
  call ShowInfo c,offset(sGood1),sGood2-sGood1
; exit termination process
  call DosExitList c,3,0)
endp ProcessComplete

dataseg
xMaxPosScreen dd xMaxScreen
yMaxPosScreen dd yMaxScreen

dataseg
GetFingerInput db 0ECh,00h,0FFh,0FFh,81h,03h,08h,00h,8+8 dup(0EEh)

dataseg
xMaxPosFinger dd xMaxFinger
yMaxPosFinger dd yMaxFinger

dataseg
TouchDownCntr db 0

udataseg
FingerDone dd ?

codeseg
proc FingerThread c near
arg @@parameter:dword
; set time critical priority
  call DosSetPriority c,2,3,31,0
; obtain finger input data
  mov esi,offset(GetFingerInput)
label AwaitFingerInputData near
; hang in here some minimum time
  call DosSleep c,250 ; mSec
; show application polling message
  call ShowInfo c,offset(sGood3),sGood4-sGood3
label GetFingerEvent near
  mov [byte(esi+6)],8 ; reset
  call DosWrite c,[fhFinger],esi,8+8,offset(FingerDone)
  call ShowReturnCode
  test eax,eax ; success
  jnz AwaitFingerInputData
; ensure valid idreport
  mov ax,[esi+8] ; idreport
  cmp al,01h ; idreport
  jne NotFingerEvent
; process finger events
  mov al,0 ; NoButMov
  cmp ah,80h ; -finger
  je FingerToMouseEvent
  mov al,6 ; But1Down
  cmp ah,81h ; +finger
  je FingerToMouseEvent
label NotFingerEvent near
; report unknown event
  mov eax,[esi+8] ; event
  bswap eax ; debugging
  call ShowReturnCode
  jmp GetFingerEvent
label FingerToMouseEvent near
; store emulated mouse event
  mov edi,offset(PutMouseEvent)
  mov [edi],al ; mouse event
; save emulated mouse position
  mov ebx,[edi+2] ; previous
; store emulated mouse xpos
  movzx eax,[word(esi)+10]
  mul [xMaxPosScreen]
  div [xMaxPosFinger]
  mov [edi+4],ax ; xpos
; store emulated mouse ypos
  movzx eax,[word(esi)+12]
  mul [yMaxPosScreen]
  div [yMaxPosFinger]
  mov [edi+2],ax ; ypos
; prepare for double click
  sub al,al ; reset counter
  mov ah,[edi] ; mouse event
; verify very same position
  cmp ebx,[edi+2] ; steady
  jne WriteFingerEvent
; verify touch down event
  cmp ah,6 ; But1Down
  jne WriteFingerEvent
  mov al,[TouchDownCntr]
; process touch down event
  cmp al,47 ; 1st click
  jna UpdateCounter
  mov ah,0 ; NoButMov
  cmp al,63 ; end click
  jna UpdateCounter
  mov ah,6 ; But1Down
  cmp al,79 ; 2nd click
  jna UpdateCounter
  mov ah,0 ; NoButMov
label UpdateCounter near
  cmp al,80 ; end click
  jnb WriteFingerEvent
  inc al ; next position
label WriteFingerEvent near
  mov [TouchDownCntr],al
  mov [edi],ah ; mouse event
; write finger event as emulated mouse event
  call DosWrite c,[fhScreen],edi,10,offset(FingerDone)
label EndWriteFingerEvent near
  jmp GetFingerEvent
endp FingerThread

codeseg
proc ProcessArguments near
; scan for forward slash
  mov al,[edi] ; character
  inc edi ; next position
  cmp al,00h ; terminator
  je EndScanString ; done
  cmp al,'/' ; parameter
  jne ProcessArguments
; verbose argument
  cmp [byte(edi)],'v'
  jne ProcessArguments
  mov [verbose],1 ; yes
  jmp ProcessArguments
label EndScanString near
; skip sanity checks
  ret ; return
endp ProcessArguments

dataseg
; verbose argument
verbose db 0 ; no

codeseg
proc ShowInfo c near
arg @@MsgOffset,@@MsgSize
; verify verbose activated
  cmp [verbose],1 ; yes
  jne EndShowInfo ; no
; show appropriate info message
  call DosWrite c,1,[@@MsgOffset],[@@MsgSize],offset(BytesDone)
label EndShowInfo near
  ret ; return
endp ShowInfo

dataseg
hex2ascii db '0123456789ABCDEF'

dataseg
szStatus db '[????????]',13,10

codeseg
proc ShowReturnCode near
; skip zero return code
  test eax,eax ; zero
  jz EndInfoShow ; done
; verify verbose activated
  cmp [verbose],1 ; yes
  jne NotInfoShow ; no
  push eax ; save register
; convert return code
  mov ecx,8 ; code length
label ConvertDigit near
  mov edx,eax ; error code
  and edx,0000000Fh ; digit
  mov dl,[hex2ascii+edx]
  mov [szStatus+ecx],dl
  shr eax,4 ; next one
  loop ConvertDigit
; show appropriate info message
  call DosWrite c,1,offset(szStatus),12,offset(BytesDone)
  pop eax ; restore register
label NotInfoShow near
  test eax,eax ; recheck
label EndInfoShow near
  ret ; return
endp ShowReturnCode

end MainRoutine
